
#include "win.h"
#include "flowstone.h"

enum {
	// Connectors for GetSysInfo
	OUT_PAGESIZE = 0,
	OUT_ALLOCSIZE = 1,

	// Connectors for OpenMapVirtual
	IN_MAPNAME = 0,
	IN_MAPSIZE = 1,
	OUT_MAPADDRESS = 2,

	// Connectors for WriteFrame
	IN_FRAME = 0,
	IN_MIX = 1,

	// Connectors for ReadFrame
	OUT_FRAME = 0,
};

// DLL Global variables
HANDLE hMutex = nullptr;
HANDLE hMapObject = nullptr;
float* lpMem = nullptr;
static int memSize = 0;

static void MakeMutex(const char* name)
{
	if (hMutex != nullptr)
	{
		CloseHandle(hMutex);
	}
	hMutex = CreateMutexA(nullptr, FALSE, name);
}

static char* MakeMutexName(const char* mappingName)
{
	static const char mutexSuffix[] = "Mutex";
	size_t sizeTotal = strlen(mappingName) + sizeof(mutexSuffix);
	char* nameMem = (char*)HeapAlloc(GetProcessHeap(), 0, sizeTotal);
	if(nameMem == nullptr)
		return nullptr;
	strcpy(nameMem, mappingName);
	strcat(nameMem, mutexSuffix);
	return nameMem;
}

static void CloseFileMapping()
{
	// If a mapping is already open, close it
	if (lpMem != nullptr)
	{
		// Unmap the view
		UnmapViewOfFile(lpMem);
		lpMem = nullptr;
	}

	if (hMapObject != nullptr)
	{
		// Close the mapping
		CloseHandle(hMapObject);
		hMapObject = nullptr;
	}

	if (hMutex != nullptr)
	{
		// Close the mutex guarding access to this mapping
		CloseHandle(hMutex);
		hMutex = nullptr;
	}
}

extern "C" __declspec(dllexport)
void __cdecl GetSysInfo(int nParams, const FSValue* pIn, FSValue* pOut)
{
	if (pIn == nullptr || pOut == nullptr || nParams < 2)
		return;

	SYSTEM_INFO siSysInfo;

	// Get Windows system information for file and memory boundaries
	GetSystemInfo(&siSysInfo);
		
	// Send to outputs
	pOut[OUT_PAGESIZE].i = siSysInfo.dwPageSize;
	pOut[OUT_ALLOCSIZE].i = siSysInfo.dwAllocationGranularity;
}

extern "C" __declspec(dllexport)
void __cdecl OpenMapVirtual(int nParams, const FSValue* pIn, FSValue* pOut)
{
	if (pIn == nullptr || pOut == nullptr || nParams < 3)
		return;

	const char* mapName = pIn[IN_MAPNAME].str;
	const int nameLen = mapName ? strlen(mapName) : 0;
	const int mapSize = pIn[IN_MAPSIZE].i;

	CloseFileMapping();

	// Attempt to open the mapped file.  hMapObject is NULL if fails
	if (nameLen > 0 && mapSize > 0)
	{
		char* mutexName = MakeMutexName(mapName);
		if (mutexName == nullptr)
			return;
		MakeMutex(mutexName);
		HeapFree(GetProcessHeap(), 0, mutexName);
		if (hMutex == nullptr)
			return;

		hMapObject = CreateFileMappingA(
			INVALID_HANDLE_VALUE, // mapping backed by pagefile
			nullptr,
			PAGE_READWRITE,
			0,
			mapSize,
			mapName);
		
		// Check if this was the first time the map was opened
		bool firstToOpen = false;

		if (hMapObject != nullptr)
		{
			firstToOpen = (GetLastError() != ERROR_ALREADY_EXISTS);

			// There is a mapping, so create a view into it
			lpMem = (float*)MapViewOfFile(
				hMapObject,
				FILE_MAP_WRITE,
				0,
				0,
				0);
		}

		if (lpMem != nullptr)
		{
			// A view was created
			if (!firstToOpen)
			{
				MEMORY_BASIC_INFORMATION meminfo {};
				if (VirtualQuery(lpMem, &meminfo, sizeof(meminfo)) > 0)
					memSize = meminfo.RegionSize;
				else
					memSize = 0;
			}
			else
			{
				memSize = mapSize;
			}
		}
		else
		{
			if (hMapObject != nullptr)
			{
				CloseHandle(hMapObject);
				hMapObject = nullptr;
			}
			memSize = 0;
		}
	}

	pOut[OUT_MAPADDRESS].i = reinterpret_cast<int>(lpMem);
}


extern "C" __declspec(dllexport)
void __cdecl CloseMap(int, const FSValue*, FSValue*)
{
	CloseFileMapping();
}

extern "C" __declspec(dllexport)
void __cdecl WriteFrame(int nParams, FSValue* pIn, FSValue* pOut)
{
	if (pIn == nullptr || pOut == nullptr || nParams < 1)
		return;

	if (lpMem == nullptr)
		return;

	if (hMutex == nullptr)
		return;

	FSFrame* frame = pIn[IN_FRAME].frame;
	if (frame == nullptr)
		return;

	const int size = min(memSize / sizeof(float), (size_t)frame->size);
	if (size <= 0)
		return;

	if (WaitForSingleObject(hMutex, INFINITE) != WAIT_OBJECT_0)
		return;

	const float* frameData = &frame->data[0];
	float* memData = lpMem;

	for (int i = 0 ; i < size ; i++)
	{
		memData[i] = frameData[i];
	}

	ReleaseMutex(hMutex);
}

extern "C" __declspec(dllexport)
void __cdecl ReadFrame(int nParams, FSValue* pIn, FSValue* pOut)
{
	if (pIn == nullptr || pOut == nullptr || nParams < 2)
		return;

	if (lpMem == nullptr)
		return;

	if (hMutex == nullptr)
		return;

	FSFrame* frame = pOut[OUT_FRAME].frame;
	if (frame == nullptr)
		return;

	int size = min(memSize / sizeof(float), (size_t)frame->size);
	if (size <= 0)
		return;

	float* frameData = &frame->data[0];
	const float* memData = lpMem;
	bool mix = pIn[IN_MIX].b;

	if (WaitForSingleObject(hMutex, INFINITE) != WAIT_OBJECT_0)
		return;

	if (mix)
	{
		for (int i = 0 ; i < size ; i++)
		{
			frameData[i] = memData[i] + frameData[i];
		}
	}
	else
	{
		for (int i = 0 ; i < size ; i++)
		{
			frameData[i] = memData[i];
		}
	}

	ReleaseMutex(hMutex);
}
